structure EDTyping :> EDTYPING =
struct

  open EDMinML

  exception TypeError


  datatype class = WellFormed | Typed of typ


  (* Make sure two types are the same. *)

  fun equal (t,t') = if t = t' then () else raise TypeError



  (* The empty environment. *)

  val empty = fn _ => raise TypeError

  type (''a, 'b) env = ''a -> 'b

  (* ++ : (''a, 'b) env * (''a * 'b) -> (''a, 'b) env 
     g++(x,t) adds a variable x with type t to environment g. *)

  infix 1 ++
  fun (g ++ (x,t)) = fn x' => if x = x' then t else g(x')



  (* typing : (string, class) env * exp -> class
     typing(g,e) returns the type of e in context g, or raises
     TypeError if e is not well-formed in context g. *)

  fun typing (_,Int _) = Typed INT
    | typing (_,Bool _) = Typed BOOL

    | typing (g,If(e,e1,e2)) =
    (checkwf (g,e);
     case (typing (g,e1),typing (g,e2)) of
       (Typed t1,Typed t2) => if t1 = t2 then Typed t1
			      else WellFormed
     | _ => WellFormed)

    | typing (g,Primop(p,es)) =
    let 
      val _ = app (fn e => (checkwf (g,e))) es
      val (args,res) = typeOfPrimop p
    in
      if length args = length es then Typed res
      else raise TypeError
    end

    | typing (g,UTFun(f,x,e)) =
    (checkwf (g++(f,WellFormed)++(x,WellFormed),e); WellFormed)
    | typing (g,Fun(f,x,t,t',e)) =
    (case typing(g++(f,Typed (ARROW(t,t')))++(x,Typed t),e) of
       Typed t'' => (equal (t',t''); Typed (ARROW(t,t')))
     | _ => raise TypeError)

    | typing (g,Apply(e,e')) =
       (case typing(g,e) of 
	  Typed (ARROW(t,t')) => (checkwf (g,e'); Typed t')
	| _ => (checkwf (g,e'); WellFormed))

    | typing (g,Var(x)) = g(x)

    | typing (g,Check (e,t)) =
	  (checkwf (g,e); Typed t)

  and checkwf (g,e) = (typing (g,e); ())

  fun typecheck e = typing(empty,e)
  fun classOpt e = SOME (typecheck e) handle TypeError => NONE

end
