structure TMinML :> TMINML =
struct

  datatype typ =
      INT
    | BOOL
    | TAGGED
    | ARROW of typ * typ

  datatype primop = Equal | Plus | Minus | Times | Negate

  datatype exp =
      Int of int

    | Bool of bool
    | If of exp * exp * exp

    | Primop of primop * exp list

    | Fun of string * string * typ * typ * exp
    | Apply of exp * exp

    | Var of string

    | TagInt of exp | TagBool of exp | TagFun of exp
    | AsInt of exp | AsBool of exp | AsFun of exp



  (* Argument and result types of all the primops. *)

  fun typeOfPrimop Equal  = ([INT, INT], BOOL)
    | typeOfPrimop Plus   = ([INT, INT], INT)
    | typeOfPrimop Minus  = ([INT, INT], INT)
    | typeOfPrimop Times  = ([INT, INT], INT)
    | typeOfPrimop Negate = ([INT], INT)

  (* Evaluation for all the primops. *)

  fun evalPrimop (Equal, [Int i, Int i']) = SOME (Bool (i = i'))
    | evalPrimop (Plus, [Int i, Int i']) = SOME (Int (i + i'))
    | evalPrimop (Minus, [Int i, Int i']) = SOME (Int (i - i'))
    | evalPrimop (Times, [Int i, Int i']) = SOME (Int (i * i'))
    | evalPrimop (Negate, [Int i]) = SOME (Int (~i))
    | evalPrimop _ = NONE

end
