structure TTyping :> TTYPING =
struct

  open TMinML

  exception TypeError



  (* Make sure two types are the same. *)

  fun equal (t,t') = if t = t' then () else raise TypeError



  (* The empty environment. *)

  val empty = fn _ => raise TypeError

  type (''a, 'b) env = ''a -> 'b

  (* ++ : (''a, 'b) env * (''a * 'b) -> (''a, 'b) env 
     g++(x,t) adds a variable x with type t to environment g. *)

  infix 1 ++
  fun (g ++ (x,t)) = fn x' => if x = x' then t else g(x')



  (* typing : (string, typ) env * exp -> typ
     typing(g,e) returns the type of e in context g, or raises
     TypeError if e has no type in context g. *)

  fun typing (_,Int _) = INT
    | typing (_,Bool _) = BOOL

    | typing (g,If(e,e1,e2)) =
      let val (t,t1,t2) = (typing(g,e), typing(g,e1), typing(g,e2))
       in equal(t,BOOL); equal(t1,t2); t1 end

    | typing (g,Primop(p,es)) =
      let val ts = map (fn e => typing(g,e)) es
          val (args,res) = typeOfPrimop p

          fun eqpairs (nil, nil) = res
            | eqpairs (h1::t1, h2::t2) = (equal (h1, h2); eqpairs (t1, t2))
            | eqpairs _ = raise TypeError
      in 
           eqpairs (ts, args)
      end

    | typing (g,Fun(f,x,t,t',e)) =
      let val t'' = typing(g++(f,ARROW(t,t'))++(x,t),e)
       in equal(t',t''); ARROW(t,t') end

    | typing (g,Apply(e,e')) =
          (case (typing(g,e),typing(g,e'))
             of (ARROW(t,t'),t'') => (equal(t,t''); t')
              | _ => raise TypeError)

    | typing (g,Var(x)) = g(x)

    | typing (g,TagInt(e)) = (equal (typing (g,e),INT);TAGGED)
    | typing (g,TagBool(e)) = (equal (typing (g,e),BOOL);TAGGED)
    | typing (g,TagFun(e)) = 
      (case typing (g,e) of ARROW _ => TAGGED | _ => raise TypeError)
    | typing (g,AsInt(e)) = (equal (typing (g,e),TAGGED);INT)
    | typing (g,AsBool(e)) = (equal (typing (g,e),TAGGED);BOOL)
    | typing (g,AsFun(e)) = (equal (typing (g,e),TAGGED);ARROW(TAGGED,TAGGED))

  fun typeOf e = typing(empty,e)
  fun typeOpt e = SOME (typeOf e) handle TypeError => NONE

end
